from species.core.species_init import SpeciesInit

__author__ = "Tomas Stolker"
__license__ = "MIT"
__version__ = "0.8.1"
__maintainer__ = "Tomas Stolker"
__email__ = "stolker@strw.leidenuniv.nl"
__status__ = "Development"
